// Copyright 2017 The Johns Hopkins University Applied Physics Laboratory.
// Licensed under the MIT License. See LICENSE.txt in the project root for full license information.



#include <eopcc/las_io.hpp>

#include <liblas/liblas.hpp>
#include <liblas/point.hpp>
#include <liblas/reader.hpp>
#include <liblas/writer.hpp>
#include <liblas/color.hpp>

#include <eopcc/cloud_io.hpp>


namespace eopcc {


  LASReader::LASReader(const std::string& filename) {
    if (!liblas::Open(ifs_, filename)) {
        throw std::runtime_error("Failed to open " + filename);
    }

    liblas::ReaderFactory factory;
    reader_ = std::unique_ptr<liblas::Reader>(new liblas::Reader(factory.CreateWithStream(ifs_)));
    liblas::Header const& header = reader_->GetHeader();


    ptformat_ = header.GetDataFormatId();
    liblas::SpatialReference sr = header.GetSRS();
    #ifdef WIN32
    // Temporary hack for Windows
    seed_cloud_.wkt_spatial_ref = ""; //sr.GetWKT();
    #else
    seed_cloud_.wkt_spatial_ref = sr.GetWKT();
    #endif

    if (ptformat_ >= kLasIncludesRGB) {
        seed_cloud_.has_color = true;
    } else {
        seed_cloud_.has_color = false;
    }

    seed_cloud_.has_las_metadata = true;

    num_points_remaining_ = header.GetPointRecordsCount();
    seed_cloud_.extent.min_x = header.GetMinX();
    seed_cloud_.extent.max_x = header.GetMaxX();

    seed_cloud_.extent.min_y = header.GetMinY();
    seed_cloud_.extent.max_y = header.GetMaxY();

    seed_cloud_.extent.min_z = header.GetMinZ();
    seed_cloud_.extent.max_z = header.GetMaxZ();
  }

  LASReader::~LASReader() {
  }

  PointCloud LASReader::ReadSubCloud(int sub_cloud_size) {
    int num_to_read = std::min<int>(sub_cloud_size, num_points_remaining_);

    PointCloud point_cloud = seed_cloud_;
    point_cloud.resize(num_to_read);
    for (size_t index=0; index < num_to_read; ++index) {
        if(!reader_->ReadNextPoint()) {
            throw std::runtime_error("Failed to read point");
        }
        num_points_remaining_ -= 1;
        liblas::Point const& point = reader_->GetPoint();

        if (ptformat_ % 2 == 1) {
            point_cloud.gps_time[index] = point.GetTime();
        }

        point_cloud.loc[index].x = point.GetX();
        point_cloud.loc[index].y = point.GetY();
        point_cloud.loc[index].z = point.GetZ();

        if (ptformat_ >= kLasIncludesRGB) {
            liblas::Color color = point.GetColor();

            point_cloud.color[index].red = color[0] / 65536.0;
            point_cloud.color[index].green = color[1] / 65536.0;
            point_cloud.color[index].blue = color[2] / 65536.0;
        }

        point_cloud.intensity[index] = point.GetIntensity() / 65536.0;

        point_cloud.las_metadata[index].return_num = point.GetReturnNumber();
        point_cloud.las_metadata[index].num_returns = point.GetNumberOfReturns();
        point_cloud.las_metadata[index].edge_of_flight_line = point.GetFlightLineEdge();
    }

    return point_cloud;
  }

  void LoadFromLAS(const std::string& las_file_name, PointCloud& point_cloud) {
    LASReader reader(las_file_name);
    point_cloud = reader.ReadSubCloud(reader.NumRemainingPoints());
  }

void TransferColors(std::string point_source, std::string color_source, std::string output_name) {
    std::ifstream ifs_color;
    if (!liblas::Open(ifs_color, color_source)) {
        throw std::runtime_error("Failed to open " + color_source);
    }

    std::ifstream ifs_point;
    if (!liblas::Open(ifs_point, point_source)) {
        throw std::runtime_error("Failed to open " + point_source);
    }

    std::ofstream ofs;
    if (!liblas::Create(ofs, output_name))
    {
        throw std::runtime_error(std::string("Can not create ") + output_name);
    }

    liblas::ReaderFactory factory;
    liblas::Reader color_reader = factory.CreateWithStream(ifs_color);
    liblas::Reader point_reader = factory.CreateWithStream(ifs_point);
    liblas::Header const& header = point_reader.GetHeader();

    liblas::Writer writer(ofs, header);
    while (color_reader.ReadNextPoint()) {
        liblas::Point const& color_point = color_reader.GetPoint();
        liblas::Point point = point_reader.GetPoint();

        while(point.GetTime() < color_point.GetTime() && point_reader.ReadNextPoint()) {
            point = point_reader.GetPoint();
        }

        point.SetColor(color_point.GetColor());
        writer.WritePoint(point);
    }
}


/// Write a PointCloud out as a LAS/LAZ file

void WriteToLAS(const std::string& las_file_name, const PointCloud& point_cloud) {
    std::ofstream ofs;
    if (!liblas::Create(ofs, las_file_name))
    {
        throw std::runtime_error(std::string("Can not create ") + las_file_name);
    }

    Extent extent = point_cloud.calculate_extent();

    liblas::Header header;

    if (las_file_name.back() == 'z') {
        header.SetCompressed(true);
    } else {
        header.SetCompressed(false);
    }

    header.SetMax(extent.max_x, extent.max_y, extent.max_z);
    header.SetMin(extent.min_x, extent.min_y, extent.min_z);

    double scale_x = ((extent.max_x - extent.min_x)/2.0) / std::numeric_limits<int32_t>::max();
    double scale_y = ((extent.max_y - extent.min_y)/2.0) / std::numeric_limits<int32_t>::max();
    double scale_z = ((extent.max_z - extent.min_z)/2.0) / std::numeric_limits<int32_t>::max();

    double offset_x = (extent.max_x - extent.min_x)/2.0 + extent.min_x;
    double offset_y = (extent.max_y - extent.min_y)/2.0 + extent.min_y;
    double offset_z = (extent.max_z - extent.min_z)/2.0 + extent.min_z;

    header.SetScale(scale_x, scale_y, scale_z);
    header.SetOffset(offset_x, offset_y, offset_z);

    header.SetVersionMajor(1);
    header.SetVersionMinor(2);
    if (point_cloud.has_color) {
        if (point_cloud.has_time) {
            header.SetDataFormatId(liblas::ePointFormat3);
        } else {
            header.SetDataFormatId(liblas::ePointFormat2);
        }
    } else {
        if (point_cloud.has_time) {
            header.SetDataFormatId(liblas::ePointFormat1);
        } else {
            header.SetDataFormatId(liblas::ePointFormat0);
        }
    }
    header.SetPointRecordsCount(point_cloud.size());
    #ifndef WIN32
    liblas::SpatialReference spatial_ref;
    spatial_ref.SetWKT(point_cloud.wkt_spatial_ref);
    header.SetSRS(spatial_ref);
    header.SetGeoreference();
    #endif

    liblas::Writer writer(ofs, header);

    for (size_t i=0; i < point_cloud.size(); ++i) {
        liblas::Point point(&header);
        point.SetCoordinates(point_cloud.loc[i].x, point_cloud.loc[i].y, point_cloud.loc[i].z);
        if (point_cloud.has_time) {
            point.SetTime(point_cloud.gps_time[i]);
        }

        if (point_cloud.has_color) {
            liblas::Color color;
            color[0] = to_idx(point_cloud.color[i].red, std::numeric_limits<uint16_t>::max() - 1);
            color[1] = to_idx(point_cloud.color[i].green, std::numeric_limits<uint16_t>::max() - 1);
            color[2] = to_idx(point_cloud.color[i].blue, std::numeric_limits<uint16_t>::max() - 1);
            point.SetColor(color);
        }


        point.SetIntensity(point_cloud.intensity[i] * 65635);

        if (point_cloud.has_las_metadata) {
          point.SetReturnNumber(point_cloud.las_metadata[i].return_num);
          point.SetNumberOfReturns(point_cloud.las_metadata[i].num_returns);
          point.SetFlightLineEdge(point_cloud.las_metadata[i].edge_of_flight_line);
        }

        writer.WritePoint(point);
    }

    writer.WriteHeader();
}




} // end eopcc
